/*
 * Copy me if you can.
 * by 20h
 */

#include <u.h>
#include <libc.h>
#include <bio.h>
#include <ip.h>
#include <ndb.h>

int primary;

static void
setdefroute(char *net, char *gate)
{
	int fd;
	char path[128];

	snprint(path, sizeof path, "%s/iproute", net);
	fd = open(path, ORDWR);
	if(fd < 0)
		return;
	fprint(fd, "add 0 0 %s", gate);
	close(fd);
}

static int
nipifcs(char *net)
{
	static Ipifc *ifc;
	Ipifc *nifc;
	Iplifc *lifc;
	int n;

	n = 0;
	ifc = readipifc(net, ifc, -1);
	for(nifc = ifc; nifc != nil; nifc = nifc->next)
		for(lifc = ifc->lifc; lifc != nil; lifc = lifc->next)
			n++;
	return n;
}

static void
putndb(char *net, char *local, char *rem)
{
	char buf[1024];
	char file[64];
	char *p, *e;
	int fd;

	e = buf + sizeof(buf);
	p = buf;
	p = seprint(p, e, "ip=%s ipmask=255.255.255.255 ipgw=%s\n", local,
			rem);
	snprint(file, sizeof file, "%s/ndb", net);
	fd = open(file, OWRITE);
	if(fd < 0)
		return;
	write(fd, buf, p-buf);
	close(fd);
}

static int
validv4(char *addr)
{

	return memcmp(addr, v4prefix, IPv4off) == 0 && memcmp(addr, v4prefix, IPaddrlen) != 0;
}

int
slipipifc(char *net, char *local, char *rem, char *mtu)
{
	int n, cfd, fd;
	char path[128];
	char buf[128];

	snprint(path, sizeof(path), "%s/ipifc/clone", net);
	cfd = open(path, ORDWR);
	if(cfd < 0)
		return -1;

	n = read(cfd, buf, sizeof(buf) - 1);
	if(n <= 0) {
		close(cfd);
		return -1;
	}
	buf[n] = 0;

	snprint(path, sizeof path, "%s/ipifc/%s/data", net, buf);
	fd = open(path, ORDWR);
	if(fd < 0) {
		close(cfd);
		return -1;
	}

	if(fprint(cfd, "bind pkt") < 0)
		return -1;
	if(fprint(cfd, "add %s 255.255.255.255 %s %s", local,
		rem, mtu) < 0) {
		close(cfd);
		return -1;
	}
	
	n = nipifcs(net);
	if(n == 0 || primary) {
		setdefroute(net, rem);
		putndb(net, local, rem);
	}

	return fd;
}

void
send_packet(int fd, char *p, int len)
{

	write(fd, "\xC0", 1);
	while(len--) {
		switch(*p & 0xFF) {
			case 0xC0:
				write(fd, "\xDBDC", 2);
				break;
			case 0xDB:
				write(fd, "\xDBDD", 2);
				break;
			default:
				write(fd, p, 1);
				break;
		}
		p++;
	}
	write(fd, "\xC0", 1);

	return;
}

int
recv_packet(int fd, char *p, int len)
{
	char c;
	int rec;

	rec = 0;

	while(1) {
		if(read(fd, &c, 1) < 0)
			return -1;
		switch(c & 0xFF) {
			case 0xC0:
				if(rec)
					return rec;
				break;
			case 0xDB:
				if(read(fd, &c, 1) < 0)
					return -1;
				switch(c & 0xFF) {
					case 0xDC:
						c = 0xC0;
						break;
					case 0xDD:
						c = 0xDB;
						break;
				}
			default:
				if(rec < len) {
					p[rec++] = c & 0xFF;
				}
		}
	}

	return 0;
}

void
dev_loop(int fd, int ipfd, char *mtu)
{
	char *b;
	int l, le;

	le = atoi(mtu);

	b = malloc(le);
	if(b == nil)
		exits("malloc");

	while(1) {
		memset(b, 0, le);
		l = recv_packet(fd, b, le);
		if(l < 0)
			goto end_dev;

		if(write(ipfd, b, le) < 0)
			goto end_dev;
	}

end_dev:
	close(ipfd);
	close(fd);
	free(b);

	exits(0);
}

void
ipifc_loop(int ipfd, int mout, char *mtu)
{
	char *b;
	int l, le;

	le = atoi(mtu);

	b = malloc(le);
	if(b == nil)
		exits("malloc");

	while(1) {
		memset(b, 0, le);
		l = read(ipfd, b, le);
		if(l < 0)
			goto end_ipfd;

		send_packet(mout, b, l);
	}

end_ipfd:
	close(mout);
	close(ipfd);
	free(b);

	exits(0);
}

void
sliprun(int min, int mout, int ipfd, char *mtu)
{

	if(rfork(RFNOWAIT|RFNOTEG|RFFDG|RFREND|RFPROC) == 0) {
		dev_loop(min, ipfd, mtu);
		exits(0);
	}

	if(rfork(RFNOWAIT|RFNOTEG|RFFDG|RFREND|RFPROC) == 0) {
		ipifc_loop(ipfd, mout, mtu);
		exits(0);
	}
		
	exits(0);
}

void
usage(void)
{

	print("usage: slip [-wP] [-b baud] [-m mtu] [-d dev] [-x netmntpt] local remote\n");
	exits(0);
}

void
main(int argc, char *argv[])
{
	int cfd, mediain, mediaout, fdipifc, windows;
	char buf[128];
	char *net, *rem, *baud, *dev, *local, *mtu;

	mtu = "1006";
	baud = "115200";
	dev = nil;
	net = "/net";
	windows = 0;

	ARGBEGIN {
		case 'b':
			baud = EARGF(usage());
			break;
		case 'm':
			mtu = EARGF(usage());
			break;
		case 'p':
			dev = EARGF(usage());
			break;
		case 'x':
			net = EARGF(usage());
			break;
		case 'w':
			windows = 1;
			break;
		case 'P':
			primary = 1;
			break;
		default:
			usage();
	} ARGEND;

	if(argc < 2)
		usage();

	local = argv[0];
	rem = argv[1];

	if(dev != nil) {
		mediain = open(dev, ORDWR);
		if(mediain < 0) {
			if(strchr(dev, '!')) {
				if((mediain = dial(dev, 0, 0, &cfd)) == -1){
					fprint(2, "slip: couldn't dial %s: %r\n", dev);
					exits(dev);
				}
			} else {
				fprint(2, "slip: couldn't open %s\n", dev);
				exits(dev);
			}
		} else {
			snprint(buf, sizeof(buf), "%sctl", dev);
			cfd = open(buf, ORDWR);
		}

		if(cfd > 0) {
			if(baud)
				fprint(cfd, "b%s", baud);
			fprint(cfd, "n1");	/* nonblocking writes on */
			close(cfd);
		}
		mediaout = mediain;
	} else {
		mediain = open("/fd/0", OREAD);
		if(mediain < 0) {
			fprint(2, "slip: couldn't open /fd/0\n");
			exits("/fd/0");
		}

		mediaout = open("/fd/1", OWRITE);
		if(mediaout < 0) {
			fprint(2, "slip: couldn't open /fd/0\n");
			exits("/fd/1");
		}
	}

	if(windows) {
		read(mediain, buf, 6);
		write(mediaout, "CLIENTSERVER", 12);
	}

	fdipifc = slipipifc(net, local, rem, mtu);
	if(fdipifc < 0)
		sysfatal("Unable to open ipifc: %r");
	sliprun(mediain, mediaout, fdipifc, mtu);

	exits(0);
}
